#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <curses.h>
#include <signal.h>

struct ut_state
{
	char tiles[9][9];
	char boards[3][3];
	int playBoard;
	char player;
};
const struct ut_state ut_initial = {
		.boards = {0},
		.tiles = {0},
		.playBoard = -1,
		.player = 'X'
	};


char ut_turn(char player)
{
	switch(player)
	{
		case 'X': return 'O';
		case 'O': return 'X';
		default: return '\0';
	}
}

#define T(r, c) (tiles[offset + stride * r + c])
char ut_winner(const char *tiles, int offset, int stride) {

	// whoreizontal wins
	for (int y = 0; y < 3; y++)
	{
		char tile = T(y, 0);
		if (tile == '\0') {continue;}
		if (tile == T(y, 1) && tile == T(y, 2)) {return tile;}
	}

	// vertical wins
	for (int x = 0; x < 3; x++)
	{
		char tile = T(0, x);
		if (tile == '\0') {continue;}
		if (tile == T(1, x) && tile == T(2, x)) {return tile;}
	}

	// diagonalz
	char tile = T(1, 1);
	if (tile != '\0')
	{
		if (tile == T(0, 0) && tile == T(2, 2)) {return tile;}
		if (tile == T(2, 0) && tile == T(0, 2)) {return tile;}
	}

	for(int y = 0; y < 3; y++)
	{
		for(int x = 0; x < 3; x++)
		{
			if(T(y, x) == '\0') {return '\0';} // in progress
		}
	}
	return ' '; // draw
}

int ut_move(struct ut_state *new_state, const struct ut_state *old_state, int row, int col)
{
	// bad move - out of bounds
	if(row < 0 || row >= 9 || col < 0 || col >= 9) {return 1;}

	// state->playBoard == -1 means all boards playable; otherwise only one board is playable
	if (old_state->playBoard != -1 && old_state->playBoard != 3 * (row / 3) + (col / 3)) {return 1;}

	// bad move - tile is occupied
	if(old_state->tiles[row][col] != '\0') {return 1;}

	// bad move - board is finished
	if(old_state->boards[row / 3][col / 3] != '\0') {return 1;}

	// copy old_state->{tiles,boards} to new_state->{tiles,boards}
	memmove(new_state->tiles, old_state->tiles, sizeof(old_state->tiles));
	memmove(new_state->boards, old_state->boards, sizeof(old_state->boards));

	// do move
	new_state->tiles[row][col] = old_state->player;
	new_state->boards[row / 3][col / 3] = ut_winner((char *)new_state->tiles, 27 * (row / 3) + 3 * (col / 3), 9);

	//  next play board - if next play board is not playable, play any board
	new_state->playBoard = new_state->boards[row % 3][col % 3] == '\0' ? 3 * (row % 3) + (col % 3) : -1;

	new_state->player = ut_turn(old_state->player);
	return 0;
}

void ut_drawBoard(const char *tiles, int offset, int stride, int iy, int ix, char winner, bool highlight)
{
	// 5x5 board display
	/*
	 * X|X|X
	 * -+-+-
	 * X|X|X
	 * -+-+-
	 * X|X|X
	*/
	for(int r = 0; r < 3; r++)
		for(int c = 0; c < 3; c++)
			mvaddch(iy + 2 * r, ix + 2 * c, T(r, c) != '\0' ? T(r, c) : ' ');
	//char info = winner != '\0' ? winner : highlight ? '*' : ' ';
	for(int r = 0; r < 3; r++)
		for(int c = 0; c < 2; c++)
			mvaddch(iy + 2 * r, ix + 2 * c + 1, winner != '\0' ? winner : '|');
	for(int r = 0; r < 2; r++)
		for(int c = 0; c < 3; c++)
			mvaddch(iy + 2 * r + 1, ix + 2 * c, winner != '\0' ? winner : '-');
	for(int r = 0; r < 2; r++)
		for(int c = 0; c < 2; c++)
			mvaddch(iy + 2 * r + 1, ix + 2 * c + 1, winner != '\0' ? winner : highlight ? '*' : '+');
}

void ut_draw(const struct ut_state *state)
{
	for(int r = 0; r < 3; r++)
		for(int c = 0; c < 3; c++)
			ut_drawBoard((char *)state->tiles,
			    27 * r + 3 * c, 9,
			    7 * r + 1, 7 * c + 1,
			    state->boards[r][c],
			    state->playBoard == -1 || state->playBoard == 3 * r + c);
	//refresh();
}

int ut_click(const struct ut_state *state, int *r, int *c)
{
#define P(r, c, d) ((state->playBoard == -1 || state->playBoard == 3 * (r / 3) + (c / 3)) && \
	          (state->boards[r / 3][c / 3] == '\0' || d % 3 == 1))
	MEVENT event;
	mvaddstr(21, 0, "Select move with arrow keys or mouse.");
	if(state->playBoard == -1)
	{
		*r = 3 * (*r / 3) + 1;
		*c = 3 * (*c / 3) + 1;
	}
	else
	{
		*r = 3 * (state->playBoard / 3) + 1;
		*c = 3 * (state->playBoard % 3) + 1;
	}
	for(;;)
	{
		/*if(state->boards[*r / 3][*c / 3] != '\0')
		{
			*r = 3 * (*r / 3) + 1;
			*c = 3 * (*c / 3) + 1;
		}*/
		move((7 * (*r / 3) + 1) + (2 * (*r % 3)),
		     (7 * (*c / 3) + 1) + (2 * (*c % 3)));
		refresh();
		switch(getch())
		{
			case KEY_MOUSE:
				if(getmouse(&event) == OK && (event.bstate & BUTTON1_CLICKED))
				{
					if(event.y < 0 || event.y >= 21) {break;}
					if(event.x < 0 || event.x >= 21) {break;}
					*r = (event.y % 7 - 1) / 2 + 3 * (event.y / 7);
					*c = (event.x % 7 - 1) / 2 + 3 * (event.x / 7);
					if(((event.y % 7) - 1) % 2 != 0) {break;}
					if(((event.x % 7) - 1) % 2 != 0) {break;}
					return 0;
				}
				break;
			case 'k':  /* FALLTHROUGH */
			case 'w':  /* FALLTHROUGH */
			case KEY_UP:
				//*r = (((*r - 1) % 9) + 9) % 9;
				for(int i = 0; i < 9 + 1; i++)
				{
					*r = (*r + 9 - 1) % 9;
					if(P(*r, *c, *r)) {break;}
				}
				break;
			case 'h':  /* FALLTHROUGH */
			case 'a':  /* FALLTHROUGH */
			case KEY_LEFT:
				//*c = (((*c - 1) % 9) + 9) % 9;
				for(int i = 0; i < 9 + 1; i++)
				{
					*c = (*c + 9 - 1) % 9;
					if(P(*r, *c, *c)) {break;}
				}
				break;
			case 'j':  /* FALLTHROUGH */
			case 's':  /* FALLTHROUGH */
			case KEY_DOWN:
				//*r = (*r + 1) % 9;
				for(int i = 0; i < 9 + 1; i++)
				{
					*r = (*r + 1) % 9;
					if(P(*r, *c, *r)) {break;}
				}
				break;
			case 'l':  /* FALLTHROUGH */
			case 'd':  /* FALLTHROUGH */
			case KEY_RIGHT:
				//*c = (*c + 1) % 9;
				for(int i = 0; i < 9 + 1; i++)
				{
					*c = (*c + 1) % 9;
					if(P(*r, *c, *c)) {break;}
				}
				break;
			case ' ':  /* FALLTHROUGH */
			case '\r': /* FALLTHROUGH */
			case KEY_ENTER:
				return 0;
			case ERR:  /* FALLTHROUGH */
			default:
				break;
		}
	}
#undef P
}

void ut_show(const struct ut_state *state) {
    #define tiles state->tiles
    printf("Turn: %c\nPlay board: %d\n", (int)state->player, state->playBoard);
	int play_board_row = -1;
	int play_board_col = -1;
	if (state->playBoard != -1) {
		play_board_row = state->playBoard / 3;
		play_board_col = state->playBoard % 3;
	}
	printf("  012 345 678\n");
    for(int y = 0; y < 9; y++)
    {
		if (y == 3 || y == 6) {
			printf("  ---+---+---\n");
		}
		printf("%d ", y);
        for(int x = 0; x < 9; x++)
        {
			if (x == 3 || x == 6) {
				putchar('|');
			}
			putchar(tiles[y][x] ? tiles[y][x] : ' ');
        }
		if (y / 3 == play_board_row) {
			printf("<");
		}
        putchar('\n');
    }
	if (play_board_col != -1) {
		printf("  ");
		for (int i = 0; i < play_board_col; i++) {
			printf("    ");
		}
		printf("^^^");
		putchar('\n');
	}
    putchar('\n');
    #undef tiles
}

bool getpos(const struct ut_state *state, int *x, int *y) {
	while (true) {
		printf("Place token %c at position x,y: ", state->player);
		char line[5] = {0};
		if (fgets(line, 5, stdin) == NULL)
			return false;
		// line was too short
		if (line[3] == '\0')
			continue;
		// line was too long
		if (line[3] != '\n') {
			// consume rest of line
			while (true) {
				int garbage = getchar();
				if (garbage == EOF)
					return false;
				if (garbage == '\n')
					break;
			}
			continue;
		}
		int sscanf_result = sscanf(line, "%d,%d", x, y);
		if (sscanf_result == EOF)
			return false;
		if (sscanf_result != 2)
			continue;
		break;
	}
	return true;
}

void finish(int sig)
{
	endwin();
	// other cleanup
	exit(0);
}

int main(int argc, char *argv[]) {
	signal(SIGINT, finish);
	initscr();
	keypad(stdscr, TRUE);
	nonl(); // \r instead of \r\n
	cbreak();
	noecho();
	mousemask(BUTTON1_CLICKED, NULL);

	struct ut_state state = ut_initial;
	while (true) {
		int r, c, err;
		erase();
		ut_draw(&state);
		char w = ut_winner((char *)state.boards, 0, 3);
		if(w)
		{
			if(w == ' ')
			{
				mvaddstr(21, 0, "Draw!");
			}
			else
			{
				mvaddch(21, 0, w);
				mvaddstr(21, 1, " Wins!");
			}
			mvaddstr(22, 0, "Press any key to exit.");
			refresh();
			mousemask(0, NULL);
			getch();
			break;
		}
		err = ut_click(&state, &r, &c);
		if (err) {continue;}
		err = ut_move(&state, &state, r, c);
		if (err) {continue;}
	}
	finish(0);
	return 0;
}
